package supply_kit_example;


import simudyne.core.abm.Action;
import simudyne.core.abm.supplykit.Facility;
import simudyne.core.abm.supplykit.Product;
import simudyne.core.abm.supplykit.utils.TransportMessages;

import java.util.*;
import java.util.stream.Collectors;


public class FacilityTypeA extends Facility<SupplyChainDemo.Globals> {

    public Map<Long, List<Product>> processedProducts = new HashMap<>();

    public static Action<FacilityTypeA> processOutboundLoadingBay =
            Action.create(FacilityTypeA.class, FacilityTypeA::processOutboundLoadingBay);

    public static Action<FacilityTypeA> departOutboundTransport =
            Action.create(FacilityTypeA.class, FacilityTypeA::departTransport);

    @Override
    protected void processOutboundLoadingBay() {
        // pull trucks from transport queue
        pullTrucksFromQueue();
        // generate products
        long processingTime = getContext().getTick() + getPrng().getNextInt(60);
        processedProducts.computeIfAbsent(processingTime, k -> new ArrayList<>()).addAll(generateProducts());
        if (processedProducts.containsKey(getContext().getTick())) {// TODO: less than current time
            List<Product> readyToDepart = processedProducts.entrySet().stream()
                    .filter(entry -> entry.getKey() <= getContext().getTick())
                    .flatMap(entry -> entry.getValue().stream())
                    .collect(Collectors.toList());
            if (readyToDepart != null) {
                readyToDepart.stream().forEach(product -> {
                    String destination = "W";
                    Optional<TransportMessages.TransportMessage> potentialTransport = outboundLoadingBay.loadingDocks.values().stream().filter(transportMessage -> transportMessage != null && transportMessage.destination == destination).findAny();
                    if (potentialTransport.isPresent()) {
                        potentialTransport.get().objectContents.add(product);
                    } else {
                        generateTruck("W");
                        pullTrucksFromQueue();
                        Optional<TransportMessages.TransportMessage> t = outboundLoadingBay.loadingDocks.values().stream().filter(transportMessage -> transportMessage != null && transportMessage.destination == destination).findAny();
                        t.get().objectContents.add(product);
                    }
                });
            }
        }
    }

    @Override
    protected void departTransport() {
        List<TransportMessages.TransportMessage> departedTransport = new ArrayList<>();
        outboundLoadingBay.loadingDocks.values().stream().filter(transport -> transport != null && transport.departureTime == getContext().getTick()).forEach(t -> {
            departedTransport.add(t);
            send(TransportMessages.TransportMessage.class, tMsg -> tMsg.constructWithProducts(t.id,
                    t.transportType, t.weightCapacity, t.volumeCapacity,
                    t.weight, t.volume, t.objectContents, t.arrivalTime,
                    t.departureTime, t.origin, t.destination
            )).to(getGlobals().destinationAddresses.get(t.destination));
        });
        outboundLoadingBay.loadingDocks.entrySet().stream().forEach(bay -> {
            if (departedTransport.contains(bay.getValue())) {
                bay.setValue(null);
            }
        });
    }

    @Override
    protected void receiveInboundTransport() {
        super.receiveInboundTransport();

    }

    @Override
    protected void processInboundLoadingBay() {

    }

    // Placeholder function for how products/units enter the system
    public List<Product> generateProducts() {
        List<Product> newProducts = new ArrayList<>();
        int nbNewProds = getPrng().getNextInt(100);
        for (int i = 0; i < nbNewProds; i++) {
            double iWeight = getPrng().uniform(10, 1000).sample();
            double iVolume = getPrng().uniform(1, 100).sample();
            String prodID = String.valueOf(getPrng().getNextInt(10000)) + "_" + getID();
            String origin = "facility_A";
            String destination = "facility_B";
            Product newProduct = new Product(prodID, iVolume, iWeight, "");
            newProducts.add(newProduct);
        }
        return newProducts;
    }

    // Placeholder function for how trucks arrive at facility
    public void generateTruck(String newDestination) {
        String trailerID = getID() + "_" + getPrng().getNextInt(1000) + "_" + newDestination;
        double weightCapacity = getPrng().uniform(10000, 25000).sample();
        double volumeCapacity = getPrng().uniform(1000, 1200).sample();
        long arrivalTime = getContext().getTick();
        long departureTime = arrivalTime + (long) getPrng().uniform(35, 100).sample();

        TransportMessages.TransportMessage transportMessage = new TransportMessages.TransportMessage();
        transportMessage.constructWithProducts(trailerID,
                "truck",
                weightCapacity,
                volumeCapacity,
                0,
                0,
                new ArrayList<>(),
                arrivalTime,
                departureTime,
                name,
                newDestination);
        outboundLoadingBay.transportQueue.put(getContext().getTick(), transportMessage);
    }

    // placeholder function for handling transport queue
    public void pullTrucksFromQueue() {
        if (outboundLoadingBay.transportQueue.size() != 0) {
            List<Integer> emptyBays = findEmptyBays(outboundLoadingBay.loadingDocks);
            while (emptyBays.size() > 0 && outboundLoadingBay.transportQueue.size() > 0) {
                long nextQueuePosition = Collections.min(outboundLoadingBay.transportQueue.keySet());
                TransportMessages.TransportMessage nextTruck = outboundLoadingBay.transportQueue.get(nextQueuePosition);
                int emptyBay = emptyBays.remove(0);
                outboundLoadingBay.loadingDocks.put(emptyBay, nextTruck);
                outboundLoadingBay.transportQueue.remove(nextQueuePosition);

            }
        }
    }

    // Placeholder function for finding empty bays
    public List<Integer> findEmptyBays(Map<Integer, TransportMessages.TransportMessage> loadingDocks) {
        List<Integer> emptyBays = new ArrayList<>();
        for (Map.Entry<Integer, TransportMessages.TransportMessage> bay : loadingDocks.entrySet()) {
            if (bay.getValue() == null) {
                emptyBays.add(bay.getKey());
            }
        }
        return emptyBays;
    }


}
