package supply_kit_example;



import java.util.*;
import java.util.stream.Collectors;

import simudyne.core.abm.Action;
import simudyne.core.abm.supplykit.*;
import simudyne.core.abm.supplykit.utils.TransportMessages;
import simudyne.core.annotations.Variable;

public class FacilityTypeB extends Facility<SupplyChainDemo.Globals> {

    @Variable
    public int nbUnitsProcessed;

    @Variable
    public double totalShipmentCost;

    public double shipmentCostPeLb = 2.5; //placeholder value for shipping cost

    public Map<String, List<Product>> inventory = new HashMap<>();

    public static Action<FacilityTypeB> receiveInboundTransport =
            Action.create(FacilityTypeB.class, FacilityTypeB::receiveInboundTransport);

    public static Action<FacilityTypeB> processInboundLoadingBay =
            Action.create(FacilityTypeB.class, FacilityTypeB::processInboundLoadingBay);

    @Override
    protected void receiveInboundTransport() {
        super.receiveInboundTransport();
    }

    @Override
    protected void processInboundLoadingBay() {
        pullArrivingTransport();
        unloadTransportCargo();
    }

    @Override
    protected void processOutboundLoadingBay() {

    }


    @Override
    protected void departTransport() {

    }

    // placeholder function for handling transport queue
    public void pullTrucksFromQueue() {
        if (inboundLoadingBay.transportQueue.size() != 0) {
            List<Integer> emptyBays = findEmptyBays(inboundLoadingBay.loadingDocks);
            while (emptyBays.size() > 0 && inboundLoadingBay.transportQueue.size() > 0) {
                long nextQueuePosition = Collections.min(inboundLoadingBay.transportQueue.keySet());
                TransportMessages.TransportMessage nextTruck = inboundLoadingBay.transportQueue.get(nextQueuePosition);
                int emptyBay = emptyBays.remove(0);
                inboundLoadingBay.transportQueue.remove(nextQueuePosition);
                inboundLoadingBay.loadingDocks.put(emptyBay, nextTruck);
            }
        }
    }

    public List<Integer> findEmptyBays(Map<Integer, TransportMessages.TransportMessage> loadingDocks) {
        List<Integer> emptyBays = new ArrayList<>();
        for (Map.Entry<Integer, TransportMessages.TransportMessage> bay : loadingDocks.entrySet()) {
            if (bay.getValue() == null) {
                emptyBays.add(bay.getKey());
            }
        }
        return emptyBays;
    }

    public void pullArrivingTransport() {
        List<TransportMessages.TransportMessage> arrivedTransport = inboundLoadingBay.transportQueue.entrySet()
                .stream()
                .filter(entry -> entry.getKey() <= getContext().getTick())
                .map(Map.Entry::getValue)
                .collect(Collectors.toList());

        if (!arrivedTransport.isEmpty()) {
            List<Integer> emptyBays = findEmptyBays(inboundLoadingBay.loadingDocks);
            while (arrivedTransport.size() > 0 && emptyBays.size() > 0) {
                TransportMessages.TransportMessage t = arrivedTransport.remove(0);
                int emptyBay = emptyBays.remove(0);
                inboundLoadingBay.loadingDocks.put(emptyBay, t);
            }
        }
    }

    public void unloadTransportCargo() {
        List<TransportMessages.TransportMessage> processedtrucks = new ArrayList<>();
        inboundLoadingBay.loadingDocks.values().stream().filter(dock -> dock != null).forEach(t -> {
            List<Product> cargo = t.objectContents;
            cargo.stream().forEach(p -> {
                nbUnitsProcessed++;
                totalShipmentCost = p.weight * shipmentCostPeLb;
                inventory.computeIfAbsent(p.productID, k -> new ArrayList<>()).add(p);
            });
            processedtrucks.add(t);
        });
        inboundLoadingBay.loadingDocks.entrySet().stream().forEach(bay -> {
            if (processedtrucks.contains(bay.getValue())) {
                bay.setValue(null);
            }
        });
    }
}
