package supply_kit_example;

import simudyne.core.abm.AgentBasedModel;
import simudyne.core.abm.GlobalState;
import simudyne.core.abm.Group;
import simudyne.core.abm.supplykit.utils.LoadingBay;
import simudyne.core.annotations.ModelSettings;
import simudyne.core.data.CSVSource;

import java.util.HashMap;
import java.util.Map;

@ModelSettings(end = 100)
public class SupplyChainDemo extends AgentBasedModel<SupplyChainDemo.Globals> {

    public static final class Globals extends GlobalState {
        public Map<String, Long> destinationAddresses;
    }

    @Override
    public void init() {
        registerAgentTypes(FacilityTypeA.class, FacilityTypeB.class);
        registerLinkTypes(Links.AToBLink.class);
    }

    @Override
    public void setup() {
        CSVSource facilityTypeASource = new CSVSource("data/facility_A_source.csv");
        Map<String, Long> destAddr = new HashMap<>();
        Group<FacilityTypeA> facilityTypeAGroup = loadGroup(FacilityTypeA.class, facilityTypeASource, fA -> {
            fA.outboundLoadingBay = new LoadingBay(2);
            destAddr.put(fA.name, fA.getID());
        });

        CSVSource facilityTypeBSource = new CSVSource("data/facility_B_source.csv");
        Group<FacilityTypeB> facilityTypeBGroup = loadGroup(FacilityTypeB.class, facilityTypeBSource, fB -> {
            fB.inboundLoadingBay = new LoadingBay(2);
            destAddr.put(fB.name, fB.getID());
        });

        getGlobals().destinationAddresses = destAddr;

        CSVSource topology = new CSVSource("data/topology.csv");
        facilityTypeAGroup.loadConnections(facilityTypeBGroup, Links.AToBLink.class, topology);

        super.setup();
    }

    @Override
    public void step() {
        super.step();
        run(FacilityTypeA.processOutboundLoadingBay);
        run(FacilityTypeA.departOutboundTransport, FacilityTypeB.receiveInboundTransport);
        run(FacilityTypeB.processInboundLoadingBay);
    }
}
